# Generated from PippCSParser.g4 by ANTLR 4.13.0
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,28,92,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,1,0,3,0,16,8,0,1,0,1,0,5,0,20,8,0,10,0,12,0,23,9,0,1,0,1,0,1,1,
        1,1,1,1,1,1,1,1,1,1,1,1,3,1,34,8,1,1,1,4,1,37,8,1,11,1,12,1,38,1,
        2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,4,2,49,8,2,11,2,12,2,50,1,3,1,3,1,
        3,1,3,1,3,1,3,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,
        4,4,4,72,8,4,11,4,12,4,73,1,4,1,4,1,4,1,5,1,5,1,5,1,5,4,5,83,8,5,
        11,5,12,5,84,1,5,1,5,1,5,1,6,1,6,1,6,0,0,7,0,2,4,6,8,10,12,0,3,2,
        0,5,5,21,21,1,1,18,18,2,0,12,13,16,16,92,0,15,1,0,0,0,2,26,1,0,0,
        0,4,40,1,0,0,0,6,52,1,0,0,0,8,58,1,0,0,0,10,78,1,0,0,0,12,89,1,0,
        0,0,14,16,5,18,0,0,15,14,1,0,0,0,15,16,1,0,0,0,16,21,1,0,0,0,17,
        20,3,2,1,0,18,20,5,18,0,0,19,17,1,0,0,0,19,18,1,0,0,0,20,23,1,0,
        0,0,21,19,1,0,0,0,21,22,1,0,0,0,22,24,1,0,0,0,23,21,1,0,0,0,24,25,
        5,0,0,1,25,1,1,0,0,0,26,27,5,1,0,0,27,28,5,2,0,0,28,29,5,18,0,0,
        29,30,5,3,0,0,30,31,5,12,0,0,31,33,5,18,0,0,32,34,3,4,2,0,33,32,
        1,0,0,0,33,34,1,0,0,0,34,36,1,0,0,0,35,37,3,8,4,0,36,35,1,0,0,0,
        37,38,1,0,0,0,38,36,1,0,0,0,38,39,1,0,0,0,39,3,1,0,0,0,40,41,5,4,
        0,0,41,42,5,28,0,0,42,43,5,22,0,0,43,44,5,23,0,0,44,45,5,24,0,0,
        45,46,5,25,0,0,46,48,5,28,0,0,47,49,3,6,3,0,48,47,1,0,0,0,49,50,
        1,0,0,0,50,48,1,0,0,0,50,51,1,0,0,0,51,5,1,0,0,0,52,53,5,26,0,0,
        53,54,5,26,0,0,54,55,5,26,0,0,55,56,5,26,0,0,56,57,5,28,0,0,57,7,
        1,0,0,0,58,59,7,0,0,0,59,60,5,12,0,0,60,61,5,18,0,0,61,62,5,6,0,
        0,62,63,5,16,0,0,63,64,5,18,0,0,64,65,5,7,0,0,65,66,5,12,0,0,66,
        67,5,18,0,0,67,68,5,8,0,0,68,69,5,12,0,0,69,71,5,18,0,0,70,72,3,
        10,5,0,71,70,1,0,0,0,72,73,1,0,0,0,73,71,1,0,0,0,73,74,1,0,0,0,74,
        75,1,0,0,0,75,76,5,9,0,0,76,77,7,1,0,0,77,9,1,0,0,0,78,79,5,16,0,
        0,79,80,3,12,6,0,80,82,5,10,0,0,81,83,5,16,0,0,82,81,1,0,0,0,83,
        84,1,0,0,0,84,82,1,0,0,0,84,85,1,0,0,0,85,86,1,0,0,0,86,87,5,11,
        0,0,87,88,5,18,0,0,88,11,1,0,0,0,89,90,7,2,0,0,90,13,1,0,0,0,8,15,
        19,21,33,38,50,73,84
    ]

class PippCSParser ( Parser ):

    grammarFileName = "PippCSParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'SHIFT_FL_FRMT'", "'RES_SIAD'", "'FIRST_RES_IN_SEQ'", 
                     "'EXP_PEAK_PICK_TBL'", "<INVALID>", "'RES_TYPE'", "'SPIN_SYSTEM_ID'", 
                     "'HETEROGENEITY'", "'END_RES_DEF'", "'('", "')'", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "'Label'", "'Exp Par Fl'", "'Peak-Pick Fl'", 
                     "'#-Cross-Ref'" ]

    symbolicNames = [ "<INVALID>", "Shift_fl_frmt", "Res_siad", "First_res_in_seq", 
                      "Exp_peak_pick_tbl", "Res_ID", "Res_type", "Spin_system_ID", 
                      "Heterogeneity", "End_res_def", "L_paren", "R_paren", 
                      "Integer", "Float", "SHARP_COMMENT", "EXCLM_COMMENT", 
                      "Simple_name", "SPACE", "RETURN", "SECTION_COMMENT", 
                      "LINE_COMMENT", "Res_ID_", "Label", "Exp_par_fl", 
                      "Peak_pick_fl", "Cross_ref", "Simple_name_ET", "SPACE_ET", 
                      "RETURN_ET" ]

    RULE_pipp_cs = 0
    RULE_pipp_format = 1
    RULE_ext_peak_pick_tbl = 2
    RULE_ext_peak_pick_tbl_row = 3
    RULE_residue_list = 4
    RULE_shift_list = 5
    RULE_number = 6

    ruleNames =  [ "pipp_cs", "pipp_format", "ext_peak_pick_tbl", "ext_peak_pick_tbl_row", 
                   "residue_list", "shift_list", "number" ]

    EOF = Token.EOF
    Shift_fl_frmt=1
    Res_siad=2
    First_res_in_seq=3
    Exp_peak_pick_tbl=4
    Res_ID=5
    Res_type=6
    Spin_system_ID=7
    Heterogeneity=8
    End_res_def=9
    L_paren=10
    R_paren=11
    Integer=12
    Float=13
    SHARP_COMMENT=14
    EXCLM_COMMENT=15
    Simple_name=16
    SPACE=17
    RETURN=18
    SECTION_COMMENT=19
    LINE_COMMENT=20
    Res_ID_=21
    Label=22
    Exp_par_fl=23
    Peak_pick_fl=24
    Cross_ref=25
    Simple_name_ET=26
    SPACE_ET=27
    RETURN_ET=28

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.0")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Pipp_csContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(PippCSParser.EOF, 0)

        def RETURN(self, i:int=None):
            if i is None:
                return self.getTokens(PippCSParser.RETURN)
            else:
                return self.getToken(PippCSParser.RETURN, i)

        def pipp_format(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PippCSParser.Pipp_formatContext)
            else:
                return self.getTypedRuleContext(PippCSParser.Pipp_formatContext,i)


        def getRuleIndex(self):
            return PippCSParser.RULE_pipp_cs

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPipp_cs" ):
                listener.enterPipp_cs(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPipp_cs" ):
                listener.exitPipp_cs(self)




    def pipp_cs(self):

        localctx = PippCSParser.Pipp_csContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_pipp_cs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 15
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                self.state = 14
                self.match(PippCSParser.RETURN)


            self.state = 21
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==1 or _la==18:
                self.state = 19
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [1]:
                    self.state = 17
                    self.pipp_format()
                    pass
                elif token in [18]:
                    self.state = 18
                    self.match(PippCSParser.RETURN)
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 23
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 24
            self.match(PippCSParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Pipp_formatContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Shift_fl_frmt(self):
            return self.getToken(PippCSParser.Shift_fl_frmt, 0)

        def Res_siad(self):
            return self.getToken(PippCSParser.Res_siad, 0)

        def RETURN(self, i:int=None):
            if i is None:
                return self.getTokens(PippCSParser.RETURN)
            else:
                return self.getToken(PippCSParser.RETURN, i)

        def First_res_in_seq(self):
            return self.getToken(PippCSParser.First_res_in_seq, 0)

        def Integer(self):
            return self.getToken(PippCSParser.Integer, 0)

        def ext_peak_pick_tbl(self):
            return self.getTypedRuleContext(PippCSParser.Ext_peak_pick_tblContext,0)


        def residue_list(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PippCSParser.Residue_listContext)
            else:
                return self.getTypedRuleContext(PippCSParser.Residue_listContext,i)


        def getRuleIndex(self):
            return PippCSParser.RULE_pipp_format

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPipp_format" ):
                listener.enterPipp_format(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPipp_format" ):
                listener.exitPipp_format(self)




    def pipp_format(self):

        localctx = PippCSParser.Pipp_formatContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_pipp_format)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 26
            self.match(PippCSParser.Shift_fl_frmt)
            self.state = 27
            self.match(PippCSParser.Res_siad)
            self.state = 28
            self.match(PippCSParser.RETURN)
            self.state = 29
            self.match(PippCSParser.First_res_in_seq)
            self.state = 30
            self.match(PippCSParser.Integer)
            self.state = 31
            self.match(PippCSParser.RETURN)
            self.state = 33
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==4:
                self.state = 32
                self.ext_peak_pick_tbl()


            self.state = 36 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 35
                self.residue_list()
                self.state = 38 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==5 or _la==21):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Ext_peak_pick_tblContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Exp_peak_pick_tbl(self):
            return self.getToken(PippCSParser.Exp_peak_pick_tbl, 0)

        def RETURN_ET(self, i:int=None):
            if i is None:
                return self.getTokens(PippCSParser.RETURN_ET)
            else:
                return self.getToken(PippCSParser.RETURN_ET, i)

        def Label(self):
            return self.getToken(PippCSParser.Label, 0)

        def Exp_par_fl(self):
            return self.getToken(PippCSParser.Exp_par_fl, 0)

        def Peak_pick_fl(self):
            return self.getToken(PippCSParser.Peak_pick_fl, 0)

        def Cross_ref(self):
            return self.getToken(PippCSParser.Cross_ref, 0)

        def ext_peak_pick_tbl_row(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PippCSParser.Ext_peak_pick_tbl_rowContext)
            else:
                return self.getTypedRuleContext(PippCSParser.Ext_peak_pick_tbl_rowContext,i)


        def getRuleIndex(self):
            return PippCSParser.RULE_ext_peak_pick_tbl

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExt_peak_pick_tbl" ):
                listener.enterExt_peak_pick_tbl(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExt_peak_pick_tbl" ):
                listener.exitExt_peak_pick_tbl(self)




    def ext_peak_pick_tbl(self):

        localctx = PippCSParser.Ext_peak_pick_tblContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_ext_peak_pick_tbl)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 40
            self.match(PippCSParser.Exp_peak_pick_tbl)
            self.state = 41
            self.match(PippCSParser.RETURN_ET)
            self.state = 42
            self.match(PippCSParser.Label)
            self.state = 43
            self.match(PippCSParser.Exp_par_fl)
            self.state = 44
            self.match(PippCSParser.Peak_pick_fl)
            self.state = 45
            self.match(PippCSParser.Cross_ref)
            self.state = 46
            self.match(PippCSParser.RETURN_ET)
            self.state = 48 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 47
                self.ext_peak_pick_tbl_row()
                self.state = 50 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==26):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Ext_peak_pick_tbl_rowContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Simple_name_ET(self, i:int=None):
            if i is None:
                return self.getTokens(PippCSParser.Simple_name_ET)
            else:
                return self.getToken(PippCSParser.Simple_name_ET, i)

        def RETURN_ET(self):
            return self.getToken(PippCSParser.RETURN_ET, 0)

        def getRuleIndex(self):
            return PippCSParser.RULE_ext_peak_pick_tbl_row

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExt_peak_pick_tbl_row" ):
                listener.enterExt_peak_pick_tbl_row(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExt_peak_pick_tbl_row" ):
                listener.exitExt_peak_pick_tbl_row(self)




    def ext_peak_pick_tbl_row(self):

        localctx = PippCSParser.Ext_peak_pick_tbl_rowContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_ext_peak_pick_tbl_row)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 52
            self.match(PippCSParser.Simple_name_ET)
            self.state = 53
            self.match(PippCSParser.Simple_name_ET)
            self.state = 54
            self.match(PippCSParser.Simple_name_ET)
            self.state = 55
            self.match(PippCSParser.Simple_name_ET)
            self.state = 56
            self.match(PippCSParser.RETURN_ET)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Residue_listContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Integer(self, i:int=None):
            if i is None:
                return self.getTokens(PippCSParser.Integer)
            else:
                return self.getToken(PippCSParser.Integer, i)

        def RETURN(self, i:int=None):
            if i is None:
                return self.getTokens(PippCSParser.RETURN)
            else:
                return self.getToken(PippCSParser.RETURN, i)

        def Res_type(self):
            return self.getToken(PippCSParser.Res_type, 0)

        def Simple_name(self):
            return self.getToken(PippCSParser.Simple_name, 0)

        def Spin_system_ID(self):
            return self.getToken(PippCSParser.Spin_system_ID, 0)

        def Heterogeneity(self):
            return self.getToken(PippCSParser.Heterogeneity, 0)

        def End_res_def(self):
            return self.getToken(PippCSParser.End_res_def, 0)

        def Res_ID(self):
            return self.getToken(PippCSParser.Res_ID, 0)

        def Res_ID_(self):
            return self.getToken(PippCSParser.Res_ID_, 0)

        def EOF(self):
            return self.getToken(PippCSParser.EOF, 0)

        def shift_list(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PippCSParser.Shift_listContext)
            else:
                return self.getTypedRuleContext(PippCSParser.Shift_listContext,i)


        def getRuleIndex(self):
            return PippCSParser.RULE_residue_list

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterResidue_list" ):
                listener.enterResidue_list(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitResidue_list" ):
                listener.exitResidue_list(self)




    def residue_list(self):

        localctx = PippCSParser.Residue_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_residue_list)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 58
            _la = self._input.LA(1)
            if not(_la==5 or _la==21):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 59
            self.match(PippCSParser.Integer)
            self.state = 60
            self.match(PippCSParser.RETURN)
            self.state = 61
            self.match(PippCSParser.Res_type)
            self.state = 62
            self.match(PippCSParser.Simple_name)
            self.state = 63
            self.match(PippCSParser.RETURN)
            self.state = 64
            self.match(PippCSParser.Spin_system_ID)
            self.state = 65
            self.match(PippCSParser.Integer)
            self.state = 66
            self.match(PippCSParser.RETURN)
            self.state = 67
            self.match(PippCSParser.Heterogeneity)
            self.state = 68
            self.match(PippCSParser.Integer)
            self.state = 69
            self.match(PippCSParser.RETURN)
            self.state = 71 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 70
                self.shift_list()
                self.state = 73 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==16):
                    break

            self.state = 75
            self.match(PippCSParser.End_res_def)
            self.state = 76
            _la = self._input.LA(1)
            if not(_la==-1 or _la==18):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Shift_listContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Simple_name(self, i:int=None):
            if i is None:
                return self.getTokens(PippCSParser.Simple_name)
            else:
                return self.getToken(PippCSParser.Simple_name, i)

        def number(self):
            return self.getTypedRuleContext(PippCSParser.NumberContext,0)


        def L_paren(self):
            return self.getToken(PippCSParser.L_paren, 0)

        def R_paren(self):
            return self.getToken(PippCSParser.R_paren, 0)

        def RETURN(self):
            return self.getToken(PippCSParser.RETURN, 0)

        def getRuleIndex(self):
            return PippCSParser.RULE_shift_list

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterShift_list" ):
                listener.enterShift_list(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitShift_list" ):
                listener.exitShift_list(self)




    def shift_list(self):

        localctx = PippCSParser.Shift_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_shift_list)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 78
            self.match(PippCSParser.Simple_name)
            self.state = 79
            self.number()
            self.state = 80
            self.match(PippCSParser.L_paren)
            self.state = 82 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 81
                self.match(PippCSParser.Simple_name)
                self.state = 84 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==16):
                    break

            self.state = 86
            self.match(PippCSParser.R_paren)
            self.state = 87
            self.match(PippCSParser.RETURN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NumberContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Float(self):
            return self.getToken(PippCSParser.Float, 0)

        def Integer(self):
            return self.getToken(PippCSParser.Integer, 0)

        def Simple_name(self):
            return self.getToken(PippCSParser.Simple_name, 0)

        def getRuleIndex(self):
            return PippCSParser.RULE_number

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumber" ):
                listener.enterNumber(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumber" ):
                listener.exitNumber(self)




    def number(self):

        localctx = PippCSParser.NumberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_number)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 89
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 77824) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








##
# File: NmrPipeCSParserListener.py
# Date: 11-Apr-2025
#
# Updates:
""" ParserLister class for NMRPIPE CS files.
    @author: Masashi Yokochi
"""
__docformat__ = "restructuredtext en"
__author__ = "Masashi Yokochi"
__email__ = "yokochi@protein.osaka-u.ac.jp"
__license__ = "Apache License 2.0"
__version__ = "1.0.0"

import sys

from antlr4 import ParseTreeListener
from typing import IO, List, Optional

try:
    from wwpdb.utils.nmr.cs.NmrPipeCSParser import NmrPipeCSParser
    from wwpdb.utils.nmr.cs.BaseCSParserListener import BaseCSParserListener
    from wwpdb.utils.nmr.AlignUtil import (emptyValue, monDict3)
    from wwpdb.utils.nmr.ChemCompUtil import ChemCompUtil
    from wwpdb.utils.nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from wwpdb.utils.nmr.nef.NEFTranslator import NEFTranslator
except ImportError:
    from nmr.cs.NmrPipeCSParser import NmrPipeCSParser
    from nmr.cs.BaseCSParserListener import BaseCSParserListener
    from nmr.AlignUtil import (emptyValue, monDict3)
    from nmr.ChemCompUtil import ChemCompUtil
    from nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from nmr.nef.NEFTranslator import NEFTranslator


# This class defines a complete listener for a parse tree produced by NmrPipeCSParser.
class NmrPipeCSParserListener(ParseTreeListener, BaseCSParserListener):

    __first_resid = 1
    __cur_sequence = ''
    __open_sequence = False
    __has_sequence = False

    __number = None

    def __init__(self, verbose: bool = True, log: IO = sys.stdout,
                 polySeq: List[dict] = None, entityAssembly: Optional[dict] = None,
                 ccU: Optional[ChemCompUtil] = None, csStat: Optional[BMRBChemShiftStat] = None, nefT: Optional[NEFTranslator] = None,
                 reasons: Optional[dict] = None):
        super().__init__(verbose, log, polySeq, entityAssembly, ccU, csStat, nefT, reasons)

        self.file_type = 'nm-shi-npi'
        self.software_name = 'NMRPipe'

    # Enter a parse tree produced by NmrPipeCSParser#nmrpipe_cs.
    def enterNmrpipe_cs(self, ctx: NmrPipeCSParser.Nmrpipe_csContext):  # pylint: disable=unused-argument
        self.enter()

    # Exit a parse tree produced by NmrPipeCSParser#nmrpipe_cs.
    def exitNmrpipe_cs(self, ctx: NmrPipeCSParser.Nmrpipe_csContext):  # pylint: disable=unused-argument
        self.exit()

    # Enter a parse tree produced by NmrPipeCSParser#sequence.
    def enterSequence(self, ctx: NmrPipeCSParser.SequenceContext):
        if self.__has_sequence and not self.__open_sequence:
            self.__first_resid = 1
            self.__cur_sequence = ''

        if ctx.First_resid():
            self.__first_resid = int(str(ctx.Integer_DA()))

        if ctx.Sequence():
            i = 0
            while ctx.One_letter_code(i):
                self.__cur_sequence += str(ctx.One_letter_code(i))
                i += 1

        self.__open_sequence = True

        self.offset = {}

    # Exit a parse tree produced by NmrPipeCSParser#sequence.
    def exitSequence(self, ctx: NmrPipeCSParser.SequenceContext):  # pylint: disable=unused-argument
        pass

    def closeSequqnce(self):
        if not self.__open_sequence:
            return

        self.__has_sequence = len(self.__cur_sequence) > 0

        self.__open_sequence = False

        if self.__has_sequence and not self.hasPolySeq:
            self.hasPolySeq = True

            def get_comp_id(one_letter_code: str):
                return next((k for k, v in monDict3.items() if v == one_letter_code and len(k) == 3), one_letter_code)

            comp_ids = [get_comp_id(one_letter_code) for one_letter_code in self.__cur_sequence]
            seq_ids = list(range(self.__first_resid, self.__first_resid + len(comp_ids) + 1))

            self.polySeq = [{'chain_id': '1', 'seq_id': seq_ids, 'comp_id': comp_ids}]
            self.entityAssembly = {'1': {'entity_id': 1, 'auth_asym_id': '.'}}

            self.labelToAuthSeq = {}
            for ps in self.polySeq:
                chainId = ps['chain_id']
                for seqId in ps['seq_id']:
                    self.labelToAuthSeq[(chainId, seqId)] = (chainId, seqId)
            self.authToLabelSeq = {v: k for k, v in self.labelToAuthSeq.items()}

            self.chainIdSet = set(ps['chain_id'] for ps in self.polySeq)
            self.compIdSet = set()

            def is_data(array: list) -> bool:
                return not any(d in emptyValue for d in array)

            for ps in self.polySeq:
                self.compIdSet.update(set(filter(is_data, ps['comp_id'])))

            for compId in self.compIdSet:
                if compId in monDict3:
                    if len(compId) == 3:
                        self.polyPeptide = True
                    elif len(compId) == 2 and compId.startswith('D'):
                        self.polyDeoxyribonucleotide = True
                    elif len(compId) == 1:
                        self.polyRibonucleotide = True

    # Enter a parse tree produced by NmrPipeCSParser#chemical_shifts.
    def enterChemical_shifts(self, ctx: NmrPipeCSParser.Chemical_shiftsContext):  # pylint: disable=unused-argument
        self.cur_list_id = max(self.cur_list_id, 0)
        self.cur_list_id += 1

        self.chemShifts = 0

        self.cur_subtype = 'chem_shift'

        self.predictSequenceNumberOffsetByFirstResidue(None, self.__first_resid, None)

        self.closeSequqnce()

    # Exit a parse tree produced by NmrPipeCSParser#chemical_shifts.
    def exitChemical_shifts(self, ctx: NmrPipeCSParser.Chemical_shiftsContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipeCSParser#chemical_shift.
    def enterChemical_shift(self, ctx: NmrPipeCSParser.Chemical_shiftContext):  # pylint: disable=unused-argument
        pass

    # Exit a parse tree produced by NmrPipeCSParser#chemical_shift.
    def exitChemical_shift(self, ctx: NmrPipeCSParser.Chemical_shiftContext):

        index = self.chemShifts + 1

        seq_id = int(str(ctx.Integer()))
        comp_id = str(ctx.Simple_name(0))
        atom_id = str(ctx.Simple_name(1))
        value = self.__number

        if value is None:
            return

        if len(comp_id) == 1 and not self.polyRibonucleotide and not self.polyDeoxyribonucleotide and self.polyPeptide:
            comp_id = next((_comp_id for _comp_id in self.compIdSet if _comp_id in monDict3 and monDict3[_comp_id] == comp_id), comp_id)

        self.atomSelectionSets.clear()

        dstFunc = self.validateCsValue(index, value, None)

        if dstFunc is None:
            return

        _seq_id = seq_id

        auth_seq_id_map = {}

        self.predictSequenceNumberOffsetByFirstResidue(None, _seq_id, comp_id)

        if None in self.offset:
            seq_id += self.offset[None]
            auth_seq_id_map[seq_id] = _seq_id

        L = f'{seq_id}:{comp_id}:{atom_id}'

        assignment = self.extractAssignment(1, L, index, with_compid=comp_id)

        if assignment is None:
            return

        has_assignments, has_multiple_assignments = self.checkAssignment(index, assignment)

        self.addCsRow(index, dstFunc, has_assignments, has_multiple_assignments, auth_seq_id_map,
                      f'{L} -> ')

        self.chemShifts += 1

    # Enter a parse tree produced by NmrPipeCSParser#chemical_shifts_sw_segid.
    def enterChemical_shifts_sw_segid(self, ctx: NmrPipeCSParser.Chemical_shifts_sw_segidContext):  # pylint: disable=unused-argument
        self.cur_list_id = max(self.cur_list_id, 0)
        self.cur_list_id += 1

        self.chemShifts = 0

        self.cur_subtype = 'chem_shift'

        self.predictSequenceNumberOffsetByFirstResidue(None, self.__first_resid, None)

        self.closeSequqnce()

    # Exit a parse tree produced by NmrPipeCSParser#chemical_shifts_sw_segid.
    def exitChemical_shifts_sw_segid(self, ctx: NmrPipeCSParser.Chemical_shifts_sw_segidContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipeCSParser#chemical_shift_sw_segid.
    def enterChemical_shift_sw_segid(self, ctx: NmrPipeCSParser.Chemical_shift_sw_segidContext):  # pylint: disable=unused-argument
        pass

    # Exit a parse tree produced by NmrPipeCSParser#chemical_shift_sw_segid.
    def exitChemical_shift_sw_segid(self, ctx: NmrPipeCSParser.Chemical_shift_sw_segidContext):

        index = self.chemShifts + 1

        chain_id = str(ctx.Simple_name(0))
        seq_id = int(str(ctx.Integer()))
        comp_id = str(ctx.Simple_name(1))
        atom_id = str(ctx.Simple_name(2))
        value = self.__number

        if value is None:
            return

        if len(comp_id) == 1 and not self.polyRibonucleotide and not self.polyDeoxyribonucleotide and self.polyPeptide:
            comp_id = next((_comp_id for _comp_id in self.compIdSet if _comp_id in monDict3 and monDict3[_comp_id] == comp_id), comp_id)

        self.atomSelectionSets.clear()

        dstFunc = self.validateCsValue(index, value, None)

        if dstFunc is None:
            return

        _seq_id = seq_id

        auth_seq_id_map = {}

        self.predictSequenceNumberOffsetByFirstResidue(chain_id, _seq_id, comp_id)

        if chain_id in self.offset:
            seq_id = _seq_id + self.offset[chain_id]
            auth_seq_id_map[seq_id] = _seq_id

        L = f'{chain_id}:{seq_id}:{comp_id}:{atom_id}'

        assignment = self.extractAssignment(1, L, index, chain_id, comp_id)

        if assignment is None:
            return

        has_assignments, has_multiple_assignments = self.checkAssignment(index, assignment)

        self.addCsRow(index, dstFunc, has_assignments, has_multiple_assignments, auth_seq_id_map,
                      f'{L} -> ')

        self.chemShifts += 1

    # Enter a parse tree produced by NmrPipeCSParser#chemical_shifts_ew_segid.
    def enterChemical_shifts_ew_segid(self, ctx: NmrPipeCSParser.Chemical_shifts_ew_segidContext):  # pylint: disable=unused-argument
        self.cur_list_id = max(self.cur_list_id, 0)
        self.cur_list_id += 1

        self.chemShifts = 0

        self.cur_subtype = 'chem_shift'

        self.predictSequenceNumberOffsetByFirstResidue(None, self.__first_resid, None)

        self.closeSequqnce()

    # Exit a parse tree produced by NmrPipeCSParser#chemical_shifts_ew_segid.
    def exitChemical_shifts_ew_segid(self, ctx: NmrPipeCSParser.Chemical_shifts_ew_segidContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by NmrPipeCSParser#chemical_shift_ew_segid.
    def enterChemical_shift_ew_segid(self, ctx: NmrPipeCSParser.Chemical_shift_ew_segidContext):  # pylint: disable=unused-argument
        pass

    # Exit a parse tree produced by NmrPipeCSParser#chemical_shift_ew_segid.
    def exitChemical_shift_ew_segid(self, ctx: NmrPipeCSParser.Chemical_shift_ew_segidContext):

        index = self.chemShifts + 1

        seq_id = int(str(ctx.Integer()))
        comp_id = str(ctx.Simple_name(0))
        atom_id = str(ctx.Simple_name(1))
        chain_id = str(ctx.Simple_name(2))
        value = self.__number

        if value is None:
            return

        if len(comp_id) == 1 and not self.polyRibonucleotide and not self.polyDeoxyribonucleotide and self.polyPeptide:
            comp_id = next((_comp_id for _comp_id in self.compIdSet if _comp_id in monDict3 and monDict3[_comp_id] == comp_id), comp_id)

        self.atomSelectionSets.clear()

        dstFunc = self.validateCsValue(index, value, None)

        if dstFunc is None:
            return

        _seq_id = seq_id

        auth_seq_id_map = {}

        self.predictSequenceNumberOffsetByFirstResidue(chain_id, _seq_id, comp_id)

        if chain_id in self.offset:
            seq_id = _seq_id + self.offset[chain_id]
            auth_seq_id_map[seq_id] = _seq_id

        L = f'{chain_id}:{seq_id}:{comp_id}:{atom_id}'

        assignment = self.extractAssignment(1, L, index, chain_id, comp_id)

        if assignment is None:
            return

        has_assignments, has_multiple_assignments = self.checkAssignment(index, assignment)

        self.addCsRow(index, dstFunc, has_assignments, has_multiple_assignments, auth_seq_id_map,
                      f'{L} -> ')

        self.chemShifts += 1

    # Enter a parse tree produced by NmrPipeCSParser#number.
    def enterNumber(self, ctx: NmrPipeCSParser.NumberContext):
        if ctx.Float():
            self.__number = float(str(ctx.Float()))

        elif ctx.Float_DecimalComma():
            self.__number = float(str(ctx.Float_DecimalComma()).replace(',', '.', 1))

        elif ctx.Integer():
            self.__number = float(str(ctx.Integer()))

        else:
            self.__number = None

    # Exit a parse tree produced by NmrPipeCSParser#number.
    def exitNumber(self, ctx: NmrPipeCSParser.NumberContext):  # pylint: disable=unused-argument
        pass


# del NmrPipeCSParser



##
# File: BareCSParserListener.py
# Date: 08-Apr-2025
#
# Updates:
""" ParserLister class for Bare WSV/TSV/CSV CS files.
    @author: Masashi Yokochi
"""
__docformat__ = "restructuredtext en"
__author__ = "Masashi Yokochi"
__email__ = "yokochi@protein.osaka-u.ac.jp"
__license__ = "Apache License 2.0"
__version__ = "1.0.0"

import sys

from antlr4 import ParseTreeListener
from typing import IO, List, Optional

try:
    from wwpdb.utils.nmr.cs.BareCSParser import BareCSParser
    from wwpdb.utils.nmr.cs.BaseCSParserListener import (BaseCSParserListener,
                                                         CHEM_SHIFT_HALF_SPIN_NUCLEUS)
    from wwpdb.utils.nmr.ChemCompUtil import ChemCompUtil
    from wwpdb.utils.nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from wwpdb.utils.nmr.nef.NEFTranslator import NEFTranslator
except ImportError:
    from nmr.cs.BareCSParser import BareCSParser
    from nmr.cs.BaseCSParserListener import (BaseCSParserListener,
                                             CHEM_SHIFT_HALF_SPIN_NUCLEUS)
    from nmr.ChemCompUtil import ChemCompUtil
    from nmr.BMRBChemShiftStat import BMRBChemShiftStat
    from nmr.nef.NEFTranslator import NEFTranslator


# This class defines a complete listener for a parse tree produced by BareCSParser.
class BareCSParserListener(ParseTreeListener, BaseCSParserListener):

    __col_name = None
    __col_order = None

    def __init__(self, verbose: bool = True, log: IO = sys.stdout,
                 polySeq: List[dict] = None, entityAssembly: Optional[dict] = None,
                 ccU: Optional[ChemCompUtil] = None, csStat: Optional[BMRBChemShiftStat] = None, nefT: Optional[NEFTranslator] = None,
                 reasons: Optional[dict] = None):
        super().__init__(verbose, log, polySeq, entityAssembly, ccU, csStat, nefT, reasons)

        self.file_type = 'nm-shi-bar'

    # Enter a parse tree produced by BareCSParser#bare_cs.
    def enterBare_cs(self, ctx: BareCSParser.Bare_csContext):  # pylint: disable=unused-argument
        self.enter()

    # Exit a parse tree produced by BareCSParser#bare_cs.
    def exitBare_cs(self, ctx: BareCSParser.Bare_csContext):  # pylint: disable=unused-argument
        self.exit()

    # Enter a parse tree produced by BareCSParser#cs_raw_format.
    def enterCs_raw_format(self, ctx: BareCSParser.Cs_raw_formatContext):

        def is_half_spin_nuclei(atom_id: str) -> bool:
            """ Return whether nuclei of a given atom_id has a spin 1/2.
                @return: True for spin 1/2 nuclei, False otherwise
            """

            return any(nucl for nucl in CHEM_SHIFT_HALF_SPIN_NUCLEUS if atom_id.startswith(nucl))

        atom_like_names = self.csStat.getAtomLikeNameSet(1)
        cs_atom_like_names = list(filter(is_half_spin_nuclei, atom_like_names))
        cs_atom_like_names.extend(['HN', 'CO'])

        self.software_name = None

        self.__col_name = []
        self.__col_order = []

        def register_column_info(col_name):
            self.__col_name.append(col_name)
            if col_name in cs_atom_like_names or col_name.endswith('#') or col_name.endswith('#'):
                self.__col_order.append('atom_name_instance')
            elif col_name.startswith('Q') or col_name.startswith('M') and 'H' + col_name[1:].replace('%', '2').replace('#', '2') in cs_atom_like_names:
                self.__col_order.append('atom_name_instance')
            elif (col_name.endswith('%') or col_name.endswith('#')) and col_name[:-1] + '2' in cs_atom_like_names:
                self.__col_order.append('atom_name_instance')
            elif 'RES' in col_name or 'SEQ' in col_name or 'COMP' in col_name or 'GROUP' in col_name or 'LABEL' in col_name:
                if 'COMP_ID' in col_name or 'NAME' in col_name or 'TYPE' in col_name or 'GROUP' in col_name or 'LABEL' in col_name:
                    self.__col_order.append('residue_name')
                else:
                    self.__col_order.append('sequence_code')
            elif 'CHAIN' in col_name or 'ASYM' in col_name or 'ENTITY' in col_name:
                self.__col_order.append('chain_code')
            elif 'ATOM' in col_name and 'TYPE' not in col_name and 'ISOTOPE' not in col_name and 'NUM' not in col_name:
                self.__col_order.append('atom_name')
            elif 'ASS' in col_name and 'ASSIGNMENTS' not in col_name:
                self.__col_order.append('assignment')
            elif 'ID' in col_name or 'NUM' in col_name or 'INDEX' in col_name:
                self.__col_order.append('index')
            elif ('SHIFT' in col_name or 'VAL' in col_name) and 'ERR' not in col_name and 'UNCERT' not in col_name:
                self.__col_order.append('value')
            elif 'ERR' in col_name or 'UNCERT' in col_name or 'DEV' in col_name:
                self.__col_order.append('value_uncertainty')
            elif ('ATOM' in col_name and 'TYPE' in col_name) or 'ELEM' in col_name or 'NUC' in col_name:
                self.__col_order.append('element')
            elif 'ISOTOPE' in col_name:
                self.__col_order.append('isotope_number')
            elif 'COMMENT' in col_name or 'DETAIL' in col_name or 'MEMO' in col_name:
                self.__col_order.append('details')
            elif 'OCC' in col_name:
                self.__col_order.append('occupancy')
            elif 'MERIT' in col_name:
                self.__col_order.append('figure_of_merit')
            elif 'AMBIG' in col_name:
                if 'SET' not in col_name:
                    self.__col_order.append('ambiguity_code')
                else:
                    self.__col_order.append('ambiguity_set_id')
            else:
                self.__col_order.append('unknown')

        if ctx.Simple_name(0):
            i = 0
            while ctx.Simple_name(i):
                register_column_info(str(ctx.Simple_name(i)).upper())
                i += 1

        elif ctx.Double_quote_string(0):
            i = 0
            while ctx.Double_quote_string(i):
                register_column_info(str(ctx.Double_quote_string(i)).strip('"').upper())
                i += 1

        sparky_resonance_columns = ('GROUP', 'ATOM', 'NUC', 'SHIFT', 'SDEV')
        if all(col_name in self.__col_name for col_name in sparky_resonance_columns):
            self.software_name = 'SPARKY'

        self.cur_list_id = max(self.cur_list_id, 0)
        self.cur_list_id += 1

        self.cur_line_num = 0

        self.chemShifts = 0
        self.offset = {}

    # Exit a parse tree produced by BareCSParser#cs_raw_format.
    def exitCs_raw_format(self, ctx: BareCSParser.Cs_raw_formatContext):  # pylint: disable=unused-argument
        pass

    # Enter a parse tree produced by BareCSParser#cs_raw_list.
    def enterCs_raw_list(self, ctx: BareCSParser.Cs_raw_listContext):  # pylint: disable=unused-argument
        self.cur_subtype = 'chem_shift'
        self.cur_line_num += 1

    # Exit a parse tree produced by BareCSParser#cs_raw_list.
    def exitCs_raw_list(self, ctx: BareCSParser.Cs_raw_listContext):  # pylint: disable=unused-argument

        def concat_assignment(chain, seq, comp, atom):
            L = ''
            if chain is not None:
                L = chain + ':'
            if seq is not None:
                L += str(seq) + ':'
            if comp is not None:
                L += comp + ':'
            L += atom
            return L

        try:

            len_ord = len(self.__col_order)
            len_any = len(self.anySelection)

            if len_ord > len_any:
                return

            if self.chemShifts == 0 and len_ord + 1 == len_any:
                if 'sequence_code' not in self.__col_order:
                    self.__col_order.insert(0, 'sequence_code')
                elif 'atom_name_instance' not in self.__col_order and 'assignment' not in self.__col_order:
                    self.__col_order.insert(0, 'assignment')
                elif 'index' not in self.__col_order:
                    self.__col_order.insert(0, 'index')
                else:
                    self.__col_order.insert(0, 'unknown')
                self.__col_name.insert(0, 'N/A')
                len_ord += 1

            if not ((('sequence_code' in self.__col_order or 'residue_name' in self.__col_order) and 'atom_name_instance' in self.__col_order)
                    or ('value' in self.__col_order and ((('sequence_code' in self.__col_order or 'residue_name' in self.__col_order) and 'atom_name' in self.__col_order)
                                                         or ('assignment' in self.__col_order and 'atom_name' not in self.__col_order)))):
                return

            if not self.hasPolySeq:
                return

            if ('sequence_code' in self.__col_order or 'residue_name' in self.__col_order) and 'atom_name_instance' in self.__col_order:

                chain_id = seq_id = comp_id = None
                atom_ids, values, details = [], [], []

                for idx in range(len_any):
                    if idx < len_ord:
                        order = self.__col_order[idx]
                        if order == 'chain_code':
                            if isinstance(self.anySelection[idx], str):
                                chain_id = self.anySelection[idx]
                            elif isinstance(self.anySelection[idx], int):
                                chain_id = str(self.anySelection[idx])
                        elif order == 'sequence_code':
                            if isinstance(self.anySelection[idx], int):
                                seq_id = self.anySelection[idx]
                            elif isinstance(self.anySelection[idx], str):
                                comp_id = self.anySelection[idx]
                        elif order == 'residue_name':
                            if isinstance(self.anySelection[idx], str):
                                comp_id = self.anySelection[idx]
                            elif isinstance(self.anySelection[idx], int):
                                seq_id = self.anySelection[idx]
                        elif order == 'atom_name_instance':
                            if isinstance(self.anySelection[idx], float):
                                value = self.anySelection[idx]
                                if self.validateCsValue(self.cur_list_id, value, None) is not None:
                                    atom_ids.append(self.__col_name[idx])
                                    values.append(value)
                            elif isinstance(self.anySelection[idx], int):
                                value = float(self.anySelection[idx])
                                if self.validateCsValue(self.cur_list_id, value, None) is not None:
                                    atom_ids.append(self.__col_name[idx])
                                    values.append(value)
                        elif order == 'details':
                            details.append(self.anySelection[idx] if isinstance(self.anySelection[idx], str) else str(self.anySelection[idx]))
                    elif self.__col_order[len_ord - 1] == 'details':
                        details.append(self.anySelection[idx] if isinstance(self.anySelection[idx], str) else str(self.anySelection[idx]))

                if len(details) == 0:
                    details = None
                else:
                    details = ' '.join(details)

                if seq_id is None or len(atom_ids) == 0:
                    return

                _seq_id = seq_id
                auth_seq_id_map = {}

                for atom_id, value in zip(atom_ids, values):
                    self.atomSelectionSets.clear()

                    dstFunc = self.validateCsValue(self.cur_line_num, value, None)

                    if dstFunc is None:
                        continue

                    self.predictSequenceNumberOffsetByFirstResidue(chain_id, _seq_id, comp_id)

                    if chain_id in self.offset:
                        seq_id = _seq_id + self.offset[chain_id]
                        auth_seq_id_map[seq_id] = _seq_id

                    L = concat_assignment(chain_id, seq_id, comp_id, atom_id)

                    assignment = self.extractAssignment(1, L, self.cur_line_num, chain_id, comp_id)

                    if assignment is None:
                        continue

                    has_assignments, has_multiple_assignments = self.checkAssignment(self.cur_line_num, assignment)

                    self.addCsRow(self.cur_line_num, dstFunc, has_assignments, has_multiple_assignments, auth_seq_id_map,
                                  f'{L} -> ', details)

                    self.chemShifts += 1

            elif ('sequence_code' in self.__col_order or 'residue_name' in self.__col_order) and 'atom_name' in self.__col_order:

                chain_id = seq_id = comp_id = None
                atom_ids, values, value_uncertainties, occupancies, figure_of_merits, details = [], [], [], [], [], []

                for idx in range(len_any):
                    if idx < len_ord:
                        order = self.__col_order[idx]
                        if order == 'chain_code':
                            if isinstance(self.anySelection[idx], str):
                                chain_id = self.anySelection[idx]
                            elif isinstance(self.anySelection[idx], int):
                                chain_id = str(self.anySelection[idx])
                        elif order == 'sequence_code':
                            if isinstance(self.anySelection[idx], int):
                                seq_id = self.anySelection[idx]
                            elif isinstance(self.anySelection[idx], str):
                                comp_id = self.anySelection[idx]
                        elif order == 'residue_name':
                            if isinstance(self.anySelection[idx], str):
                                comp_id = self.anySelection[idx]
                            elif isinstance(self.anySelection[idx], int):
                                seq_id = self.anySelection[idx]
                        elif order == 'atom_name':
                            if isinstance(self.anySelection[idx], str):
                                atom_ids.append(self.anySelection[idx])
                        elif order == 'value':
                            value = self.anySelection[idx]
                            if isinstance(value, float):
                                if self.validateCsValue(self.cur_list_id, value, None) is not None:
                                    values.append(self.anySelection[idx])
                                elif len(atom_ids) > len(values):
                                    del atom_ids[-1]
                            elif isinstance(value, int):
                                value = float(value)
                                if self.validateCsValue(self.cur_list_id, value, None) is not None:
                                    values.append(self.anySelection[idx])
                                elif len(atom_ids) > len(values):
                                    del atom_ids[-1]
                            elif len(atom_ids) > len(values):
                                del atom_ids[-1]
                        elif order == 'value_uncertainty':
                            if len(values) - 1 == len(value_uncertainties):
                                value_uncertainty = self.anySelection[idx]
                                if isinstance(value_uncertainty, float):
                                    value_uncertainties.append(value_uncertainty)
                                else:
                                    value_uncertainties.append(None)
                        elif order == 'occupancy':
                            if len(values) - 1 == len(occupancies):
                                occupancy = self.anySelection[idx]
                                if isinstance(occupancy, float):
                                    occupancies.append(occupancy)
                                else:
                                    occupancies.append(None)
                        elif order == 'figure_of_merit':
                            if len(values) - 1 == len(figure_of_merits):
                                figure_of_merit = self.anySelection[idx]
                                if isinstance(figure_of_merit, float):
                                    figure_of_merits.append(figure_of_merit)
                                else:
                                    figure_of_merits.append(None)
                        elif order == 'details':
                            details.append(self.anySelection[idx] if isinstance(self.anySelection[idx], str) else str(self.anySelection[idx]))
                    elif self.__col_order[len_ord - 1] == 'details':
                        details.append(self.anySelection[idx] if isinstance(self.anySelection[idx], str) else str(self.anySelection[idx]))

                if len(details) == 0:
                    details = None
                else:
                    details = ' '.join(details)

                if seq_id is None or len(atom_ids) == 0:
                    return

                _seq_id = seq_id
                auth_seq_id_map = {}

                for atom_id, value, value_uncertainty, occupancy, figure_of_merit in zip(atom_ids, values, value_uncertainties, occupancies, figure_of_merits):
                    self.atomSelectionSets.clear()

                    dstFunc = self.validateCsValue(self.cur_line_num, value, value_uncertainty, occupancy, figure_of_merit)

                    if dstFunc is None:
                        continue

                    self.predictSequenceNumberOffsetByFirstResidue(chain_id, _seq_id, comp_id)

                    if chain_id in self.offset:
                        seq_id = _seq_id + self.offset[chain_id]
                        auth_seq_id_map[seq_id] = _seq_id

                    L = concat_assignment(chain_id, seq_id, comp_id, atom_id)

                    assignment = self.extractAssignment(1, L, self.cur_line_num, chain_id, comp_id)

                    if assignment is None:
                        continue

                    has_assignments, has_multiple_assignments = self.checkAssignment(self.cur_line_num, assignment)

                    self.addCsRow(self.cur_line_num, dstFunc, has_assignments, has_multiple_assignments, auth_seq_id_map,
                                  f'{L} -> ', details)

                    self.chemShifts += 1

            else:

                chain_id = None
                assignments, values, value_uncertainties, occupancies, figure_of_merits, details = [], [], [], [], [], []

                for idx in range(len_any):
                    if idx < len_ord:
                        order = self.__col_order[idx]
                        if order == 'chain_code':
                            if isinstance(self.anySelection[idx], str):
                                chain_id = self.anySelection[idx]
                            elif isinstance(self.anySelection[idx], int):
                                chain_id = str(self.anySelection[idx])
                        elif order == 'assignment':
                            if isinstance(self.anySelection[idx], str):
                                assignments.append(self.anySelection[idx])
                        elif order == 'value':
                            value = self.anySelection[idx]
                            if isinstance(value, float):
                                if self.validateCsValue(self.cur_list_id, value, None) is not None:
                                    values.append(self.anySelection[idx])
                                elif len(atom_ids) > len(values):
                                    del atom_ids[-1]
                            elif isinstance(value, int):
                                value = float(value)
                                if self.validateCsValue(self.cur_list_id, value, None) is not None:
                                    values.append(self.anySelection[idx])
                                elif len(atom_ids) > len(values):
                                    del atom_ids[-1]
                            elif len(atom_ids) > len(values):
                                del atom_ids[-1]
                        elif order == 'value_uncertainty':
                            if len(values) - 1 == len(value_uncertainties):
                                value_uncertainty = self.anySelection[idx]
                                if isinstance(value_uncertainty, float):
                                    value_uncertainties.append(value_uncertainty)
                                else:
                                    value_uncertainties.append(None)
                        elif order == 'occupancy':
                            if len(values) - 1 == len(occupancies):
                                occupancy = self.anySelection[idx]
                                if isinstance(occupancy, float):
                                    occupancies.append(occupancy)
                                else:
                                    occupancies.append(None)
                        elif order == 'figure_of_merit':
                            if len(values) - 1 == len(figure_of_merits):
                                figure_of_merit = self.anySelection[idx]
                                if isinstance(figure_of_merit, float):
                                    figure_of_merits.append(figure_of_merit)
                                else:
                                    figure_of_merits.append(None)
                        elif order == 'details':
                            details.append(self.anySelection[idx] if isinstance(self.anySelection[idx], str) else str(self.anySelection[idx]))
                    elif self.__col_order[len_ord - 1] == 'details':
                        details.append(self.anySelection[idx] if isinstance(self.anySelection[idx], str) else str(self.anySelection[idx]))

                if len(details) == 0:
                    details = None
                else:
                    details = ' '.join(details)

                if len(assignments) == 0:
                    return

                for L, value, value_uncertainty, occupancy, figure_of_merit in zip(assignments, values, value_uncertainties, occupancies, figure_of_merits):
                    self.atomSelectionSets.clear()

                    dstFunc = self.validateCsValue(self.cur_line_num, value, value_uncertainty, occupancy, figure_of_merit)

                    if dstFunc is None:
                        continue

                    if chain_id is not None:
                        L = f'{chain_id}:{L}'

                    assignment = self.extractAssignment(1, L, self.cur_line_num, chain_id)

                    if assignment is None:
                        continue

                    has_assignments, has_multiple_assignments = self.checkAssignment(self.cur_line_num, assignment)

                    self.addCsRow(self.cur_line_num, dstFunc, has_assignments, has_multiple_assignments, {},
                                  f'{L} -> ', details)

                    self.chemShifts += 1

        finally:
            self.anySelection.clear()

    # Enter a parse tree produced by BareCSParser#any.
    def enterAny(self, ctx: BareCSParser.AnyContext):

        try:

            if ctx.Float():
                value = str(ctx.Float())
                self.anySelection.append(float(value))

            elif ctx.Integer():
                value = str(ctx.Integer())
                self.anySelection.append(int(value))

            elif ctx.Simple_name():
                self.anySelection.append(str(ctx.Simple_name()))

            elif ctx.Double_quote_float():
                value = str(ctx.Double_quote_float()).strip('"')
                self.anySelection.append(float(value))

            elif ctx.Double_quote_integer():
                value = str(ctx.Double_quote_integer()).strip('"')
                self.anySelection.append(int(value))

            else:
                self.anySelection.append(str(ctx.Double_quote_string()).strip('"'))

        except ValueError:
            self.anySelection.append(None)

    # Exit a parse tree produced by BareCSParser#any.
    def exitAny(self, ctx: BareCSParser.AnyContext):  # pylint: disable=unused-argument
        pass


# del BareCSParser



# Generated from PonderosaPKParser.g4 by ANTLR 4.13.0
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,21,99,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,1,0,3,0,18,8,0,1,0,1,0,1,0,1,0,5,0,24,8,0,10,0,12,0,27,
        9,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,4,1,40,8,1,11,1,
        12,1,41,1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,3,1,3,1,3,1,3,1,3,1,3,1,3,
        1,3,1,3,4,3,60,8,3,11,3,12,3,61,1,4,1,4,1,4,1,4,1,4,1,4,1,4,1,4,
        1,4,1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,1,5,4,5,82,8,5,11,5,12,5,83,
        1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,7,1,7,1,7,0,0,8,0,
        2,4,6,8,10,12,14,0,2,1,1,11,11,1,0,4,5,98,0,17,1,0,0,0,2,30,1,0,
        0,0,4,43,1,0,0,0,6,50,1,0,0,0,8,63,1,0,0,0,10,72,1,0,0,0,12,85,1,
        0,0,0,14,96,1,0,0,0,16,18,5,11,0,0,17,16,1,0,0,0,17,18,1,0,0,0,18,
        25,1,0,0,0,19,24,3,2,1,0,20,24,3,6,3,0,21,24,3,10,5,0,22,24,5,11,
        0,0,23,19,1,0,0,0,23,20,1,0,0,0,23,21,1,0,0,0,23,22,1,0,0,0,24,27,
        1,0,0,0,25,23,1,0,0,0,25,26,1,0,0,0,26,28,1,0,0,0,27,25,1,0,0,0,
        28,29,5,0,0,1,29,1,1,0,0,0,30,31,5,1,0,0,31,32,5,14,0,0,32,33,5,
        15,0,0,33,34,5,17,0,0,34,35,5,2,0,0,35,36,5,18,0,0,36,37,5,19,0,
        0,37,39,5,21,0,0,38,40,3,4,2,0,39,38,1,0,0,0,40,41,1,0,0,0,41,39,
        1,0,0,0,41,42,1,0,0,0,42,3,1,0,0,0,43,44,5,4,0,0,44,45,5,4,0,0,45,
        46,3,14,7,0,46,47,5,9,0,0,47,48,5,9,0,0,48,49,7,0,0,0,49,5,1,0,0,
        0,50,51,5,1,0,0,51,52,5,14,0,0,52,53,5,15,0,0,53,54,5,17,0,0,54,
        55,5,2,0,0,55,56,5,18,0,0,56,57,5,19,0,0,57,59,5,21,0,0,58,60,3,
        8,4,0,59,58,1,0,0,0,60,61,1,0,0,0,61,59,1,0,0,0,61,62,1,0,0,0,62,
        7,1,0,0,0,63,64,5,4,0,0,64,65,5,4,0,0,65,66,5,4,0,0,66,67,3,14,7,
        0,67,68,5,9,0,0,68,69,5,9,0,0,69,70,5,9,0,0,70,71,7,0,0,0,71,9,1,
        0,0,0,72,73,5,1,0,0,73,74,5,14,0,0,74,75,5,15,0,0,75,76,5,17,0,0,
        76,77,5,2,0,0,77,78,5,18,0,0,78,79,5,19,0,0,79,81,5,21,0,0,80,82,
        3,12,6,0,81,80,1,0,0,0,82,83,1,0,0,0,83,81,1,0,0,0,83,84,1,0,0,0,
        84,11,1,0,0,0,85,86,5,4,0,0,86,87,5,4,0,0,87,88,5,4,0,0,88,89,5,
        4,0,0,89,90,3,14,7,0,90,91,5,9,0,0,91,92,5,9,0,0,92,93,5,9,0,0,93,
        94,5,9,0,0,94,95,7,0,0,0,95,13,1,0,0,0,96,97,7,1,0,0,97,15,1,0,0,
        0,6,17,23,25,41,61,83
    ]

class PonderosaPKParser ( Parser ):

    grammarFileName = "PonderosaPKParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'NOESYTYPE'", "'AXISORDER'" ]

    symbolicNames = [ "<INVALID>", "Noesy_type", "Axis_order", "Integer", 
                      "Float", "Real", "SHARP_COMMENT", "EXCLM_COMMENT", 
                      "SMCLN_COMMENT", "Simple_name", "SPACE", "RETURN", 
                      "SECTION_COMMENT", "LINE_COMMENT", "Integer_NT", "Simple_name_NT", 
                      "SPACE_NT", "RETURN_NT", "Integer_AO", "Simple_name_AO", 
                      "SPACE_AO", "RETURN_AO" ]

    RULE_ponderosa_pk = 0
    RULE_peak_list_2d = 1
    RULE_peak_2d = 2
    RULE_peak_list_3d = 3
    RULE_peak_3d = 4
    RULE_peak_list_4d = 5
    RULE_peak_4d = 6
    RULE_number = 7

    ruleNames =  [ "ponderosa_pk", "peak_list_2d", "peak_2d", "peak_list_3d", 
                   "peak_3d", "peak_list_4d", "peak_4d", "number" ]

    EOF = Token.EOF
    Noesy_type=1
    Axis_order=2
    Integer=3
    Float=4
    Real=5
    SHARP_COMMENT=6
    EXCLM_COMMENT=7
    SMCLN_COMMENT=8
    Simple_name=9
    SPACE=10
    RETURN=11
    SECTION_COMMENT=12
    LINE_COMMENT=13
    Integer_NT=14
    Simple_name_NT=15
    SPACE_NT=16
    RETURN_NT=17
    Integer_AO=18
    Simple_name_AO=19
    SPACE_AO=20
    RETURN_AO=21

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.0")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Ponderosa_pkContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(PonderosaPKParser.EOF, 0)

        def RETURN(self, i:int=None):
            if i is None:
                return self.getTokens(PonderosaPKParser.RETURN)
            else:
                return self.getToken(PonderosaPKParser.RETURN, i)

        def peak_list_2d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PonderosaPKParser.Peak_list_2dContext)
            else:
                return self.getTypedRuleContext(PonderosaPKParser.Peak_list_2dContext,i)


        def peak_list_3d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PonderosaPKParser.Peak_list_3dContext)
            else:
                return self.getTypedRuleContext(PonderosaPKParser.Peak_list_3dContext,i)


        def peak_list_4d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PonderosaPKParser.Peak_list_4dContext)
            else:
                return self.getTypedRuleContext(PonderosaPKParser.Peak_list_4dContext,i)


        def getRuleIndex(self):
            return PonderosaPKParser.RULE_ponderosa_pk

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPonderosa_pk" ):
                listener.enterPonderosa_pk(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPonderosa_pk" ):
                listener.exitPonderosa_pk(self)




    def ponderosa_pk(self):

        localctx = PonderosaPKParser.Ponderosa_pkContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_ponderosa_pk)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 17
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                self.state = 16
                self.match(PonderosaPKParser.RETURN)


            self.state = 25
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==1 or _la==11:
                self.state = 23
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
                if la_ == 1:
                    self.state = 19
                    self.peak_list_2d()
                    pass

                elif la_ == 2:
                    self.state = 20
                    self.peak_list_3d()
                    pass

                elif la_ == 3:
                    self.state = 21
                    self.peak_list_4d()
                    pass

                elif la_ == 4:
                    self.state = 22
                    self.match(PonderosaPKParser.RETURN)
                    pass


                self.state = 27
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 28
            self.match(PonderosaPKParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_list_2dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Noesy_type(self):
            return self.getToken(PonderosaPKParser.Noesy_type, 0)

        def Integer_NT(self):
            return self.getToken(PonderosaPKParser.Integer_NT, 0)

        def Simple_name_NT(self):
            return self.getToken(PonderosaPKParser.Simple_name_NT, 0)

        def RETURN_NT(self):
            return self.getToken(PonderosaPKParser.RETURN_NT, 0)

        def Axis_order(self):
            return self.getToken(PonderosaPKParser.Axis_order, 0)

        def Integer_AO(self):
            return self.getToken(PonderosaPKParser.Integer_AO, 0)

        def Simple_name_AO(self):
            return self.getToken(PonderosaPKParser.Simple_name_AO, 0)

        def RETURN_AO(self):
            return self.getToken(PonderosaPKParser.RETURN_AO, 0)

        def peak_2d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PonderosaPKParser.Peak_2dContext)
            else:
                return self.getTypedRuleContext(PonderosaPKParser.Peak_2dContext,i)


        def getRuleIndex(self):
            return PonderosaPKParser.RULE_peak_list_2d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_list_2d" ):
                listener.enterPeak_list_2d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_list_2d" ):
                listener.exitPeak_list_2d(self)




    def peak_list_2d(self):

        localctx = PonderosaPKParser.Peak_list_2dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_peak_list_2d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 30
            self.match(PonderosaPKParser.Noesy_type)
            self.state = 31
            self.match(PonderosaPKParser.Integer_NT)
            self.state = 32
            self.match(PonderosaPKParser.Simple_name_NT)
            self.state = 33
            self.match(PonderosaPKParser.RETURN_NT)
            self.state = 34
            self.match(PonderosaPKParser.Axis_order)
            self.state = 35
            self.match(PonderosaPKParser.Integer_AO)
            self.state = 36
            self.match(PonderosaPKParser.Simple_name_AO)
            self.state = 37
            self.match(PonderosaPKParser.RETURN_AO)
            self.state = 39 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 38
                self.peak_2d()
                self.state = 41 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==4):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_2dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(PonderosaPKParser.Float)
            else:
                return self.getToken(PonderosaPKParser.Float, i)

        def number(self):
            return self.getTypedRuleContext(PonderosaPKParser.NumberContext,0)


        def Simple_name(self, i:int=None):
            if i is None:
                return self.getTokens(PonderosaPKParser.Simple_name)
            else:
                return self.getToken(PonderosaPKParser.Simple_name, i)

        def RETURN(self):
            return self.getToken(PonderosaPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(PonderosaPKParser.EOF, 0)

        def getRuleIndex(self):
            return PonderosaPKParser.RULE_peak_2d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_2d" ):
                listener.enterPeak_2d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_2d" ):
                listener.exitPeak_2d(self)




    def peak_2d(self):

        localctx = PonderosaPKParser.Peak_2dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_peak_2d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 43
            self.match(PonderosaPKParser.Float)
            self.state = 44
            self.match(PonderosaPKParser.Float)
            self.state = 45
            self.number()
            self.state = 46
            self.match(PonderosaPKParser.Simple_name)
            self.state = 47
            self.match(PonderosaPKParser.Simple_name)
            self.state = 48
            _la = self._input.LA(1)
            if not(_la==-1 or _la==11):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_list_3dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Noesy_type(self):
            return self.getToken(PonderosaPKParser.Noesy_type, 0)

        def Integer_NT(self):
            return self.getToken(PonderosaPKParser.Integer_NT, 0)

        def Simple_name_NT(self):
            return self.getToken(PonderosaPKParser.Simple_name_NT, 0)

        def RETURN_NT(self):
            return self.getToken(PonderosaPKParser.RETURN_NT, 0)

        def Axis_order(self):
            return self.getToken(PonderosaPKParser.Axis_order, 0)

        def Integer_AO(self):
            return self.getToken(PonderosaPKParser.Integer_AO, 0)

        def Simple_name_AO(self):
            return self.getToken(PonderosaPKParser.Simple_name_AO, 0)

        def RETURN_AO(self):
            return self.getToken(PonderosaPKParser.RETURN_AO, 0)

        def peak_3d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PonderosaPKParser.Peak_3dContext)
            else:
                return self.getTypedRuleContext(PonderosaPKParser.Peak_3dContext,i)


        def getRuleIndex(self):
            return PonderosaPKParser.RULE_peak_list_3d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_list_3d" ):
                listener.enterPeak_list_3d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_list_3d" ):
                listener.exitPeak_list_3d(self)




    def peak_list_3d(self):

        localctx = PonderosaPKParser.Peak_list_3dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_peak_list_3d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 50
            self.match(PonderosaPKParser.Noesy_type)
            self.state = 51
            self.match(PonderosaPKParser.Integer_NT)
            self.state = 52
            self.match(PonderosaPKParser.Simple_name_NT)
            self.state = 53
            self.match(PonderosaPKParser.RETURN_NT)
            self.state = 54
            self.match(PonderosaPKParser.Axis_order)
            self.state = 55
            self.match(PonderosaPKParser.Integer_AO)
            self.state = 56
            self.match(PonderosaPKParser.Simple_name_AO)
            self.state = 57
            self.match(PonderosaPKParser.RETURN_AO)
            self.state = 59 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 58
                self.peak_3d()
                self.state = 61 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==4):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_3dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(PonderosaPKParser.Float)
            else:
                return self.getToken(PonderosaPKParser.Float, i)

        def number(self):
            return self.getTypedRuleContext(PonderosaPKParser.NumberContext,0)


        def Simple_name(self, i:int=None):
            if i is None:
                return self.getTokens(PonderosaPKParser.Simple_name)
            else:
                return self.getToken(PonderosaPKParser.Simple_name, i)

        def RETURN(self):
            return self.getToken(PonderosaPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(PonderosaPKParser.EOF, 0)

        def getRuleIndex(self):
            return PonderosaPKParser.RULE_peak_3d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_3d" ):
                listener.enterPeak_3d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_3d" ):
                listener.exitPeak_3d(self)




    def peak_3d(self):

        localctx = PonderosaPKParser.Peak_3dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_peak_3d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 63
            self.match(PonderosaPKParser.Float)
            self.state = 64
            self.match(PonderosaPKParser.Float)
            self.state = 65
            self.match(PonderosaPKParser.Float)
            self.state = 66
            self.number()
            self.state = 67
            self.match(PonderosaPKParser.Simple_name)
            self.state = 68
            self.match(PonderosaPKParser.Simple_name)
            self.state = 69
            self.match(PonderosaPKParser.Simple_name)
            self.state = 70
            _la = self._input.LA(1)
            if not(_la==-1 or _la==11):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_list_4dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Noesy_type(self):
            return self.getToken(PonderosaPKParser.Noesy_type, 0)

        def Integer_NT(self):
            return self.getToken(PonderosaPKParser.Integer_NT, 0)

        def Simple_name_NT(self):
            return self.getToken(PonderosaPKParser.Simple_name_NT, 0)

        def RETURN_NT(self):
            return self.getToken(PonderosaPKParser.RETURN_NT, 0)

        def Axis_order(self):
            return self.getToken(PonderosaPKParser.Axis_order, 0)

        def Integer_AO(self):
            return self.getToken(PonderosaPKParser.Integer_AO, 0)

        def Simple_name_AO(self):
            return self.getToken(PonderosaPKParser.Simple_name_AO, 0)

        def RETURN_AO(self):
            return self.getToken(PonderosaPKParser.RETURN_AO, 0)

        def peak_4d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PonderosaPKParser.Peak_4dContext)
            else:
                return self.getTypedRuleContext(PonderosaPKParser.Peak_4dContext,i)


        def getRuleIndex(self):
            return PonderosaPKParser.RULE_peak_list_4d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_list_4d" ):
                listener.enterPeak_list_4d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_list_4d" ):
                listener.exitPeak_list_4d(self)




    def peak_list_4d(self):

        localctx = PonderosaPKParser.Peak_list_4dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_peak_list_4d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 72
            self.match(PonderosaPKParser.Noesy_type)
            self.state = 73
            self.match(PonderosaPKParser.Integer_NT)
            self.state = 74
            self.match(PonderosaPKParser.Simple_name_NT)
            self.state = 75
            self.match(PonderosaPKParser.RETURN_NT)
            self.state = 76
            self.match(PonderosaPKParser.Axis_order)
            self.state = 77
            self.match(PonderosaPKParser.Integer_AO)
            self.state = 78
            self.match(PonderosaPKParser.Simple_name_AO)
            self.state = 79
            self.match(PonderosaPKParser.RETURN_AO)
            self.state = 81 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 80
                self.peak_4d()
                self.state = 83 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==4):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_4dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(PonderosaPKParser.Float)
            else:
                return self.getToken(PonderosaPKParser.Float, i)

        def number(self):
            return self.getTypedRuleContext(PonderosaPKParser.NumberContext,0)


        def Simple_name(self, i:int=None):
            if i is None:
                return self.getTokens(PonderosaPKParser.Simple_name)
            else:
                return self.getToken(PonderosaPKParser.Simple_name, i)

        def RETURN(self):
            return self.getToken(PonderosaPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(PonderosaPKParser.EOF, 0)

        def getRuleIndex(self):
            return PonderosaPKParser.RULE_peak_4d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_4d" ):
                listener.enterPeak_4d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_4d" ):
                listener.exitPeak_4d(self)




    def peak_4d(self):

        localctx = PonderosaPKParser.Peak_4dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_peak_4d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 85
            self.match(PonderosaPKParser.Float)
            self.state = 86
            self.match(PonderosaPKParser.Float)
            self.state = 87
            self.match(PonderosaPKParser.Float)
            self.state = 88
            self.match(PonderosaPKParser.Float)
            self.state = 89
            self.number()
            self.state = 90
            self.match(PonderosaPKParser.Simple_name)
            self.state = 91
            self.match(PonderosaPKParser.Simple_name)
            self.state = 92
            self.match(PonderosaPKParser.Simple_name)
            self.state = 93
            self.match(PonderosaPKParser.Simple_name)
            self.state = 94
            _la = self._input.LA(1)
            if not(_la==-1 or _la==11):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NumberContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Float(self):
            return self.getToken(PonderosaPKParser.Float, 0)

        def Real(self):
            return self.getToken(PonderosaPKParser.Real, 0)

        def getRuleIndex(self):
            return PonderosaPKParser.RULE_number

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumber" ):
                listener.enterNumber(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumber" ):
                listener.exitNumber(self)




    def number(self):

        localctx = PonderosaPKParser.NumberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_number)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 96
            _la = self._input.LA(1)
            if not(_la==4 or _la==5):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








##
# File: ChemCompReader.py
# Date: 31-May-2010  John Westbrook
#
# Update:
# 06-Aug-2010 - jdw - generalized construction of methods to apply to any category
#                     add accessors for lists of dictionaries
# 12-May-2011 - rps - added check for None when asking for category Object in __getDataList()
# 24-Oct-2012 - rps - updated to reflect reorganization of modules in pdbx packages
# 17-Jan-2025 - my  - added is_reserved_lig_code() from AlignUtil.py (DAOTHER-7204, 7388)
##
""" A collection of classes parsing CCD CIF files.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook, Masashi Yokochi"
__email__ = "jwest@rcsb.rutgers.edu, yokochi@protein.osaka-u.ac.jp"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "1.0.3"

import sys
import os
import re

from mmcif.api.PdbxContainers import DataContainer
from mmcif.io.PdbxReader import PdbxReader

from typing import IO, List, Optional

try:
    from wwpdb.utils.nmr.AlignUtil import emptyValue
except ImportError:
    from nmr.AlignUtil import emptyValue


ccd_id_pattern = re.compile(r'(\w{1,3}|\w{5})')


def is_reserved_lig_code(comp_id: str) -> bool:
    """ Return a given comp_id is reserved for new ligands. (DAOTHER-7204, 7388)
    """

    if comp_id in ('LIG', 'DRG', 'INH'):
        return True

    if len(comp_id) == 2 and comp_id[0].isdigit() and comp_id[1].isdigit() and comp_id != '00':
        return True

    return False


class ChemCompReader:
    """ Accessor methods for parsing CCD CIF files.
    """

    def __init__(self, verbose: bool = True, log: IO = sys.stdout):
        self.__class_name__ = self.__class__.__name__
        self.__version__ = __version__

        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False

        # the primary datablock
        self.__dBlock = None

        # the top file tree of CCD
        self.__topCachePath = None

        # the current file path
        self.__filePath = None

        # the current compId
        self.__compId = None

        # the compId of the current datablock
        self.__lastCompId = None

        # items of CCD category
        self.__itemDict = {
            'chem_comp': [
                ('_chem_comp.id', '%s', 'str', ''),
                ('_chem_comp.name', '%s', 'str', ''),
                ('_chem_comp.type', '%s', 'str', ''),
                ('_chem_comp.pdbx_type', '%s', 'str', ''),
                ('_chem_comp.formula', '%s', 'str', ''),
                ('_chem_comp.mon_nstd_parent_comp_id', '%s', 'str', ''),
                ('_chem_comp.pdbx_synonyms', '%s', 'str', ''),
                ('_chem_comp.pdbx_formal_charge', '%s', 'str', ''),
                ('_chem_comp.pdbx_initial_date', '%s', 'str', ''),
                ('_chem_comp.pdbx_modified_date', '%s', 'str', ''),
                ('_chem_comp.pdbx_ambiguous_flag', '%s', 'str', ''),
                ('_chem_comp.pdbx_release_status', '%s', 'str', ''),
                ('_chem_comp.pdbx_replaced_by', '%s', 'str', ''),
                ('_chem_comp.pdbx_replaces', '%s', 'str', ''),
                ('_chem_comp.formula_weight', '%s', 'str', ''),
                ('_chem_comp.one_letter_code', '%s', 'str', ''),
                ('_chem_comp.three_letter_code', '%s', 'str', ''),
                ('_chem_comp.pdbx_model_coordinates_details', '%s', 'str', ''),
                ('_chem_comp.pdbx_model_coordinates_missing_flag', '%s', 'str', ''),
                ('_chem_comp.pdbx_ideal_coordinates_details', '%s', 'str', ''),
                ('_chem_comp.pdbx_ideal_coordinates_missing_flag', '%s', 'str', ''),
                ('_chem_comp.pdbx_model_coordinates_db_code', '%s', 'str', ''),
                ('_chem_comp.pdbx_subcomponent_list', '%s', 'str', ''),
                ('_chem_comp.pdbx_processing_site', '%s', 'str', '')
            ],
            'chem_comp_atom': [
                ('_chem_comp_atom.comp_id', '%s', 'str', ''),
                ('_chem_comp_atom.atom_id', '%s', 'str', ''),
                ('_chem_comp_atom.alt_atom_id', '%s', 'str', ''),
                ('_chem_comp_atom.type_symbol', '%s', 'str', ''),
                ('_chem_comp_atom.charge', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_align', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_aromatic_flag', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_leaving_atom_flag', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_stereo_config', '%s', 'str', ''),
                ('_chem_comp_atom.model_Cartn_x', '%s', 'str', ''),
                ('_chem_comp_atom.model_Cartn_y', '%s', 'str', ''),
                ('_chem_comp_atom.model_Cartn_z', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_model_Cartn_x_ideal', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_model_Cartn_y_ideal', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_model_Cartn_z_ideal', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_component_atom_id', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_component_comp_id', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_ordinal', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_backbone_atom_flag', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_n_terminal_atom_flag', '%s', 'str', ''),
                ('_chem_comp_atom.pdbx_c_terminal_atom_flag', '%s', 'str', '')
            ],
            'chem_comp_bond': [
                ('_chem_comp_bond.comp_id', '%s', 'str', ''),
                ('_chem_comp_bond.atom_id_1', '%s', 'str', ''),
                ('_chem_comp_bond.atom_id_2', '%s', 'str', ''),
                ('_chem_comp_bond.value_order', '%s', 'str', ''),
                ('_chem_comp_bond.pdbx_aromatic_flag', '%s', 'str', ''),
                ('_chem_comp_bond.pdbx_stereo_config', '%s', 'str', ''),
                ('_chem_comp_bond.pdbx_ordinal', '%s', 'str', '')
            ],
            'chem_comp_descriptor': [
                ('_pdbx_chem_comp_descriptor.comp_id', '%s', 'str', ''),
                ('_pdbx_chem_comp_descriptor.type', '%s', 'str', ''),
                ('_pdbx_chem_comp_descriptor.program', '%s', 'str', ''),
                ('_pdbx_chem_comp_descriptor.program_version', '%s', 'str', ''),
                ('_pdbx_chem_comp_descriptor.descriptor', '%s', 'str', '')
            ],
            'chem_comp_identifier': [
                ('_pdbx_chem_comp_identifier.comp_id', '%s', 'str', ''),
                ('_pdbx_chem_comp_identifier.type', '%s', 'str', ''),
                ('_pdbx_chem_comp_identifier.program', '%s', 'str', ''),
                ('_pdbx_chem_comp_identifier.program_version', '%s', 'str', ''),
                ('_pdbx_chem_comp_identifier.identifier', '%s', 'str', '')
            ]
        }

    def setCachePath(self, topCachePath: str = '/data/components/ligand-dict-v4'):
        """ Set the top file tree of CCD.
        """

        self.__topCachePath = topCachePath

    def setCompId(self, compId: str, ligand: bool = True) -> bool:
        """ Set compId and CCD CIF file path based on file tree of CCD.
        """

        if compId in emptyValue:
            return False

        self.__compId = compId.upper()

        if not ccd_id_pattern.match(self.__compId) or (not ligand and is_reserved_lig_code(self.__compId)):
            return False

        self.__filePath = os.path.join(self.__topCachePath,
                                       self.__compId[-2:] if len(self.__compId) > 3 else self.__compId[0],
                                       self.__compId,
                                       self.__compId + '.cif')

        if not os.access(self.__filePath, os.R_OK):
            if self.__verbose:
                self.__lfh.write(f"+{self.__class_name__}.setCompId() ++ Error  - Missing file {self.__filePath}\n")
            return False

        return True

    def setFilePath(self, filePath: str, compId: str) -> bool:
        """ Set compId and CCD CIF file path.
        """

        try:

            if compId in emptyValue:
                return False

            self.__compId = compId.upper()

            if not ccd_id_pattern.match(self.__compId) or is_reserved_lig_code(self.__compId):
                return False

            self.__filePath = filePath

            if not os.access(self.__filePath, os.R_OK):
                if self.__verbose:
                    self.__lfh.write(f"+{self.__class_name__}.setFilePath() ++ Error  - Missing file {self.__filePath}\n")
                return False

            return True

        except Exception as e:
            if self.__verbose:
                self.__lfh.write(f"+{self.__class_name__}.setFilePath() ++ Error  - Set {self.__filePath} failed {str(e)}\n")
            return False

    def getAtomList(self) -> List[list]:
        """ Get a list of list of data from the chem_comp_atom category.
        """

        self.__updateDataBlock()

        return self.__getRowList(catName='chem_comp_atom')

    def getBonds(self) -> List[list]:
        """ Get a list of list of data from the chem_comp_bond category.
        """

        self.__updateDataBlock()

        return self.__getRowList(catName='chem_comp_bond')

    def getChemCompDict(self) -> dict:
        """ Get a dictionary of the chem_comp category.
        """

        try:

            self.__updateDataBlock()

            return self.__getDictList(catName='chem_comp')[0]

        except Exception:
            return {}

    def __updateDataBlock(self) -> bool:
        """ Update the definition data for the input compId.
            @return: True for success or False otherwise
        """

        if self.__compId == self.__lastCompId:
            return True

        try:

            if self.__setDataBlock(self.__getDataBlock()):
                self.__lastCompId = self.__compId
                return True

            return False

        except Exception as e:
            if self.__verbose:
                self.__lfh.write(f"+{self.__class_name__}.__updateDataBlock() ++ Error  - {str(e)}\n")
            return False

    def __getDataBlock(self, blockId: Optional[str] = None) -> Optional[DataContainer]:
        """ Worker method to read CCD CIF file and set the target datablock.
            @return: the first datablock if no blockId is provided
        """

        try:

            with open(self.__filePath, 'r', encoding='utf-8') as ifh:
                myBlockList = []
                pRd = PdbxReader(ifh)
                pRd.read(myBlockList)

            if blockId is not None:

                for block in myBlockList:
                    if (block.getType() == 'data' and block.getName() == blockId):
                        if self.__verbose and self.__debug:
                            block.printIt(self.__lfh)
                        return block

            else:

                for block in myBlockList:
                    if block.getType() == 'data':
                        if self.__verbose and self.__debug:
                            block.printIt(self.__lfh)
                        return block

        except Exception as e:
            if self.__verbose:
                self.__lfh.write(f"+{self.__class_name__}.__getDataBlock() ++ Error  - {str(e)}\n")

        return None

    def __setDataBlock(self, dataBlock: Optional[DataContainer]) -> bool:
        """ Assigns the input datablock as the active internal datablock.
        """

        if dataBlock is None:
            self.__dBlock = None
            return False

        try:

            if dataBlock.getType() == 'data':
                self.__dBlock = dataBlock
                return True

        except Exception:
            pass

        self.__dBlock = None

        return False

    def __getDictList(self, catName: str = 'chem_comp') -> List[dict]:
        """ Return a list of dictionaries of the input category.
        """

        catObj = self.__dBlock.getObj(catName)

        if catObj is None:
            return []

        itDict = {itName: idxIt for idxIt, itName in enumerate(catObj.getItemNameList())}

        colDict = {}
        for itTup in self.__itemDict[catName]:
            colDict[itTup[0]] = itDict[itTup[0]] if itTup[0] in itDict else -1

        dList = []
        for row in catObj.getRowList():
            dList.append({itName: '' if idxIt < 0 else row[idxIt] for itName, idxIt in colDict.items()})

        return dList

    def __getRowList(self, catName: str = 'chem_comp_bond') -> List[list]:
        """ Return a list a list of data from the input category including
            data types and default value replacement.
        """

        catObj = self.__dBlock.getObj(catName)

        if catObj is None:
            return []

        def apply_type(ctype, default, val):
            if val in emptyValue:
                return default
            if ctype == 'int':
                return int(val)
            if ctype == 'float':
                return float(val)
            return val

        itDict = {itName: idxIt for idxIt, itName in enumerate(catObj.getItemNameList())}

        colTupList = []
        for itTup in self.__itemDict[catName]:
            colTupList.append((itDict[itTup[0]] if itTup[0] in itDict else -1, itTup[2], itTup[3]))

        dList = []
        for row in catObj.getRowList():
            uR = []
            for cTup in colTupList:
                uR.append(apply_type(cTup[1], cTup[2], cTup[2] if cTup[0] < 0 else row[cTup[0]]))
            dList.append(uR)

        return dList
